/*
 * xzx    : Copyright (c) Des Herriott 1993, 1994
 * xcoupe : Copyright (c) Allan Skillman 1995.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Allan Skillman (Based on xzx coding by Des Herriott)
 */

/* Source file main.c - initial setup etc. */

#define Z80_MAIN

#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#include "z80.h"
#include "mem.h"
#include "xsam.h"
#include "rtcfg.h"
#include "util.h"
#include "fdi.h"

static void got_sig(int signo);
static void read_roms(void);

int
main(int argc, char **argv)
{
	int i;

	/* Make sure we can't be killed off while initialising stuff, 
	 * or we might end up leaving shared memory segments lying about.  */
	signal(SIGINT,  SIG_IGN);
	signal(SIGHUP,  SIG_IGN);
	signal(SIGTERM, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);

	start_emulator(&argc, argv);

	init_memory();
        init_ports();
	init_pages();
	read_roms();

        eject_disk(1);
        eject_disk(2);
        
	insert_disk(GetCfg(disk_1), 1);
	insert_disk(GetCfg(disk_2), 2);


	signal(SIGINT,  got_sig);
	signal(SIGHUP,  got_sig);
	signal(SIGTERM, got_sig);
	signal(SIGQUIT, got_sig);

	xcoupe_main_loop();

	return 0;
}


/* Read the ROM image <image> into real memory page <page> */
static int
load_rom_image(char *image, int page)
{
	FILE *ifp;
	size_t nread;

	if ((ifp = open_xcoupe_file(image, "r")) == NULL) {
		xcoupe_mesg(XCOUPE_ERR, "can't open ROM image <%s>", image);
		return -1;
	}

	nread = fread(realMem[page], 1, XCOUPE_PAGESIZE, ifp);

	/* give a warning if image is not 16K */
	if (nread != XCOUPE_PAGESIZE ) {
		xcoupe_mesg(XCOUPE_WARN, "ROM image <%s> is only %d bytes", image, nread);
	}

	fclose(ifp);

	return 0;
}


void
read_roms(void)
{
		load_rom_image(GetCfg(rom0), ROM0);
		load_rom_image(GetCfg(rom1), ROM1);
}


static void
got_sig(int signo)
{
	Quit(0);
}


