/*
 * xzx    : Copyright (c) Des Herriott 1993, 1994
 * xcoupe : Copyright (c) Allan Skillman 1995.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Allan Skillman
 */

/* Source file mouse.c - Emulation of the SAM mouse */

#include <stdio.h>

#include "z80.h"
#include "mouse.h"
#include "xsam.h"


/* Buffer used to store the Mouse data */

typedef union _MouseBuff {
	struct _MouseStruct {
		uns8 MouseStrobe;
		uns8 MouseDummy;
		uns8 Button_status;
		uns8 Y256;
		uns8 Y16;
		uns8 Y1;
		uns8 X256;
		uns8 X16;
		uns8 X1;
	}    MouseStruct;
	uns8 MouseArray[9]; 
} MouseBuff;

int MouseStrobed = 0;

static MouseBuff *TheMouseBuff;

static uns8 *BuffPtr;
static int  ByteCount = 0;	


void 
mouse_init()
{
	TheMouseBuff = (MouseBuff *)malloc(sizeof(MouseBuff));
	mouse_reset(1);	
}

uns8 
mouse_in()
{
	uns8 in_byte;

	if (!ByteCount)
		MouseStrobed = 1;

	if (ByteCount++ < 9) {
		in_byte = *BuffPtr++;
	}
	else {
		in_byte = 0xff;		/* dummy */
		mouse_reset(0); 
		mouse_xy_write(0,0);
		ByteCount++;		/* Skip the strobe */
		BuffPtr++;
	}
/*	printf("mouse_in()=%d\n",in_byte);	*/
	return(in_byte); 
}

void 
mouse_reset(int ResetButtons) 
{
/* Clear the counters for the mouse */

	int i;

	ByteCount = 0;
	BuffPtr = TheMouseBuff->MouseArray;

	*BuffPtr = *(BuffPtr + 1) = 0xff;
	if (ResetButtons) 
		*(BuffPtr + 2) = 0xff;

	/* There is never a need to actually clear the position
	   counters - as this is done when the pointer is warped
	   back to the center of the window - ie we always get
	   a mouse_xy_write(0,0) event                         */

}

void 
mouse_xy_write(int DeltaX, int DeltaY)
{
/* Write the X,Y change DeltaX, DeltaY into the mouse buffer */

/*	printf("mouse_xy_write(%d,%d)\n",DeltaX,DeltaY); */
	TheMouseBuff->MouseStruct.Y256 = (DeltaY & 0x0f00) >> 8;
	TheMouseBuff->MouseStruct.Y16  = (DeltaY & 0x00f0) >> 4;
	TheMouseBuff->MouseStruct.Y1   = (DeltaY & 0x000f);
	TheMouseBuff->MouseStruct.X256 = (DeltaX & 0x0f00) >> 8;
	TheMouseBuff->MouseStruct.X16  = (DeltaX & 0x00f0) >> 4;
	TheMouseBuff->MouseStruct.X1   = (DeltaX & 0x000f);
}

void 
mouse_button_write(int ButtonData)
{
/* Write the Button status into mouse buffer */

	TheMouseBuff->MouseStruct.Button_status = (uns8)ButtonData;	
}

