/*
 * xzx    : Copyright (c) Des Herriott 1993, 1994
 * xcoupe : Copyright (c) Allan Skillman 1995.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of the copyright holder not be used in
 * advertising or publicity pertaining to distribution of the software without
 * specific, written prior permission.  The copyright holder makes no
 * representations about the suitability of this software for any purpose.  It
 * is provided "as is" without express or implied warranty.
 *
 * THE COPYRIGHT HOLDER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
 * EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY SPECIAL, INDIRECT OR
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,
 * DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
 * TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 * Author: Allan Skillman (Based on xzx coding by Des Herriott)
 */

/* Source file xsam.c - emulation of SAM Coupe hardware */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef MITSHM
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
#if defined(nec_ews) && !defined(__STDC__)
void *shmat(int, void*, int);	/* shut the compiler up */
#endif
#endif

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "mem.h"
#include "xsam.h"
#include "xsamP.h"
#include "rtcfg.h"
#include "util.h"
#include "version.h"
#include "mouse.h"
#include "xspectrum.h"

#ifdef XPM_ICON
#include <X11/xpm.h>
#include "xcoupe_icon.xpm"
#else
#include "xcoupe.icon"
#endif

Display *dpy;
static Screen *sptr;
static int scrn;
static Window root;
static Colormap cmap;
static GC theGc;
static GC fgGc, bgGc;
static Window borderWin, mainWin;
KeySym *key_table;
static XImage *xim;
static Pixmap thePix;
static XFontStruct *theFont;
#ifdef MITSHM
static XShmSegmentInfo xshminfo;
#endif

void (*screen_write)(uns16, uns8);
void (*screen_write_xim_mode_1)(uns16, uns8);
void (*screen_write_xim_mode_2)(uns16, uns8);
void (*screen_write_xim_mode_3)(uns16, uns8);
void (*screen_write_xim_mode_4)(uns16, uns8);
void (*screen_refresh)(void);

uns8 keyports[10];	/* 8 rows of keys (+ 1 row for unscanned keys) */
uns8 inPorts[256], outPorts[256];

static int xpos, ypos;

/* All to do with screen painting */
enum {XVal, YVal};
extern unsigned char dtbl_x[6144];		/* maps dfile offset to X coord */
extern unsigned char dtbl_y[6144];		/* maps odfile ffset to Y coord */
static unsigned long pixels[NCOLORS];		/* 128  SAM colours */
static int screen_dirty;			/* does the screen need refreshing? */
static int x_min, x_max, y_min, y_max;		/* bounding box of last area changed */
static int bytes_per_line;			/* no. of bytes in 1 line of XImage */
static int xim_total_size;			/* Total size of Ximage (used by out_clut) */
static char *xim_data_ptr;			/* pointer to actual XImage data */		


static char *shadow_image;			/* Pointer to the shadow image used for fast
						   clut update */

static int first_line[16];			/* first scan line where a clut change 
						   is made for each colour */
static int line_int_disabled;			/* signals that the line interrupt register
						   has been set to > 191 in the frame */
int clut_change_in_frame;			/* signals a clut change has been made in 
						   the present frame */
int border_change_in_frame;			/* signals a border change has been made in
						   the present frame */
int vmpr_change_in_frame;			/* signals a vmpr change has been made in
						   the present frame */
						   
int DoPaletteSwitching;				/* Flag for palette updating on line interrupts */
int SpectrumKeyboard;				/* Switch for spectrum keyboard
						   mode - see xspectrum.c */

int ScaleFactor;				/* Mode for dynamic scale switching */
#define HSize		(HSIZE*ScaleFactor)	/* Macros used in scale switching */
#define VSize		(VSIZE*ScaleFactor)
#define BorderWidth	(BORDER_WIDTH*ScaleFactor)
static inline void out_vmpr(uns8 val);

/* These macros map a byte in the attribute area to an X11 pixel value */
#define Background(byte) (((byte) & 0x38) >> 3) | (((byte) & 0x40) >> 3)
#define Foreground(byte) ((byte) & 0x07) | (((byte) & 0x40) >> 3)
/* Map an (x,y) pixel coordinate to a offset into the attribute area */
#define XYToAttr(x,y) (((x) >> 3) + ((y & 0xf8) << 2))

/* right-hand edge of the bounding box must be rounded up */
static uns8 round_up_x[] = { 0x07, 0x0f, 0x17, 0x1f };

static int flash_phase = 0;	/* gets toggled between 1 and 0 twice per second */

#define SAMPointerX  (HSize/2)		/* Initial coordinates of the SAM */
#define SAMPointerY  (VSize/2)		/* mouse pointer */

static int SAMButtons  = 0xff;  	/* Mouse button status */
static int SAMPointerMode = 0;		/* Pointer Attatched and x magnification */
static Cursor SAMNullCursor;		/* Special Null Cursor */			

/* from resource.c */
extern Display *xcoupe_open_display(int *argcp, char **argvp, KeySym *key_table);

/* from emul.c */
extern int LineNo;
extern int disass;
extern int InterruptType;
 
/* all to do with the popup dialog */
#define EF_MAX_VISIBLE 50
#define EF_MAX 256
#define BUTTON_LABEL "OK"
#define GUTTER_HT 5
#define GUTTER_WID 5
static void make_dialog(int *argcp, char **argvp);
static Window dlg_win, entry_win, ok_win;
static int dlg_wid, dlg_ht, dlg_x, dlg_y;
static int char_width, font_ht;
static int label_wid, label_ht, entry_wid, entry_ht, ok_wid, ok_ht;
static int cursor_pos = 0, cur_win_pos = 0;
static char dlg_string[EF_MAX];

/* The SAM ports */

uns8 vmpr = 0x81;
uns8 hmpr = 0x01;
uns8 lmpr = 0x1f;
uns8 midi = 0x00;	
uns8 border = 0x00;
uns8 clut[16] = {0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 
                 0x00, 0x19, 0x2a, 0x3b, 0x4c, 0x5d, 0x6e, 0x7f};
uns8 keyboard = 0x5f;
uns8 status_reg = 0xff;
uns8 line_int = 0xff;

uns8 hpen = 0x00;
uns8 lpen = 0x00;
        	
uns8 clock_reg = 0xff;

uns8 lepr = 0x00;
uns8 hepr = 0x00;

uns8 attr_reg = 0xff;

/*
 *  Private functions start here
 */

/* Ximage Screen writing routines; two for each of the four modes, */
/* one for 1x1 and one for 2x2 scaling */

void
screen_write_xim_mode_1_1x1(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns16 offset;
	unsigned int ink, paper;
	unsigned long i_pix, p_pix, tmp;
	uns8 attr;

	uns8 pmask, ipmask;

	screen_dirty++;

	offset = addr;

	x = dtbl_x[offset]; y = dtbl_y[offset];
	attr = realMem[VMPR_SCREEN_PAGE][6144 + XYToAttr(x,y)];
	ink = Foreground(attr); paper = Background(attr);
	/* toggle the colours if we're in the inverse part of the FLASH cycle */
	if (flash_phase && (attr & 0x80)) {
		tmp=ink; ink=paper; paper=tmp;
	}

	i_pix = pixels[clut[ink]];
	p_pix = pixels[clut[paper]];

	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y;

	ByteOffset = x + y*HSize*2;
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	baseptr[0] = val & 0x80? i_pix: p_pix;
	baseptr[1] = val & 0x40? i_pix: p_pix;
	baseptr[2] = val & 0x20? i_pix: p_pix;
	baseptr[3] = val & 0x10? i_pix: p_pix;
	baseptr[4] = val & 0x08? i_pix: p_pix;
	baseptr[5] = val & 0x04? i_pix: p_pix;
	baseptr[6] = val & 0x02? i_pix: p_pix;
	baseptr[7] = val & 0x01? i_pix: p_pix;
	ShadowBaseptr[0] = val & 0x80? ink: paper;
	ShadowBaseptr[1] = val & 0x40? ink: paper;
	ShadowBaseptr[2] = val & 0x20? ink: paper;
	ShadowBaseptr[3] = val & 0x10? ink: paper;
	ShadowBaseptr[4] = val & 0x08? ink: paper;
	ShadowBaseptr[5] = val & 0x04? ink: paper;
	ShadowBaseptr[6] = val & 0x02? ink: paper;
	ShadowBaseptr[7] = val & 0x01? ink: paper;

}

void
screen_write_xim_mode_1_2x2(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns16 offset;
	unsigned int ink, paper;
	unsigned long i_pix, p_pix, tmp;
	uns8 attr;

	uns8 pmask, ipmask;
	int j, k;

	screen_dirty++;

	offset = addr;

	x = dtbl_x[offset]; y = dtbl_y[offset];
	attr = realMem[VMPR_SCREEN_PAGE][6144 + XYToAttr(x,y)];
	ink = Foreground(attr); paper = Background(attr);
	/* toggle the colours if we're in the inverse part of the FLASH cycle */
	if (flash_phase && (attr & 0x80)) {
		tmp=ink; ink=paper; paper=tmp;
	}

	i_pix = pixels[clut[ink]];
	p_pix = pixels[clut[paper]];

	x *= 2; y *= 2;

	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x + 1;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y + 1;

	ByteOffset = x + y*HSize;
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	for (j = 0; j < 2; j++)
		for (k = 0; k < 2; k++) {
		        pixel_offset = j * HSize + k; 
			baseptr[        pixel_offset] = val & 0x80? i_pix: p_pix;
			baseptr[1 * 2 + pixel_offset] = val & 0x40? i_pix: p_pix;
			baseptr[2 * 2 + pixel_offset] = val & 0x20? i_pix: p_pix;
			baseptr[3 * 2 + pixel_offset] = val & 0x10? i_pix: p_pix;
			baseptr[4 * 2 + pixel_offset] = val & 0x08? i_pix: p_pix;
			baseptr[5 * 2 + pixel_offset] = val & 0x04? i_pix: p_pix;
			baseptr[6 * 2 + pixel_offset] = val & 0x02? i_pix: p_pix;
			baseptr[7 * 2 + pixel_offset] = val & 0x01? i_pix: p_pix;
			ShadowBaseptr[        pixel_offset] = val & 0x80? ink: paper;
			ShadowBaseptr[1 * 2 + pixel_offset] = val & 0x40? ink: paper;
			ShadowBaseptr[2 * 2 + pixel_offset] = val & 0x20? ink: paper;
			ShadowBaseptr[3 * 2 + pixel_offset] = val & 0x10? ink: paper;
			ShadowBaseptr[4 * 2 + pixel_offset] = val & 0x08? ink: paper;
			ShadowBaseptr[5 * 2 + pixel_offset] = val & 0x04? ink: paper;
			ShadowBaseptr[6 * 2 + pixel_offset] = val & 0x02? ink: paper;
			ShadowBaseptr[7 * 2 + pixel_offset] = val & 0x01? ink: paper;
		}

}

void
screen_write_xim_mode_2_1x1(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns16 offset;
	unsigned int ink, paper;
	unsigned long i_pix, p_pix, tmp;
	uns8 attr;

	uns8 pmask, ipmask;

	screen_dirty++;

	offset = addr;

	y = offset >> 5;
	x = (offset & 0x1f) << 3;
	
	attr = realMem[VMPR_SCREEN_PAGE][8192 + addr];
	ink = Foreground(attr); paper = Background(attr);
	/* toggle the colours if we're in the inverse part of the FLASH cycle */
	if (flash_phase && (attr & 0x80)) {
		tmp=ink; ink=paper; paper=tmp;
	}

	i_pix = pixels[clut[ink]];
	p_pix = pixels[clut[paper]];

	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y;

	ByteOffset = x + y*HSize*2;
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	baseptr[0] = val & 0x80? i_pix: p_pix;
	baseptr[1] = val & 0x40? i_pix: p_pix;
	baseptr[2] = val & 0x20? i_pix: p_pix;
	baseptr[3] = val & 0x10? i_pix: p_pix;
	baseptr[4] = val & 0x08? i_pix: p_pix;
	baseptr[5] = val & 0x04? i_pix: p_pix;
	baseptr[6] = val & 0x02? i_pix: p_pix;
	baseptr[7] = val & 0x01? i_pix: p_pix;
	ShadowBaseptr[0] = val & 0x80? ink: paper;
	ShadowBaseptr[1] = val & 0x40? ink: paper;
	ShadowBaseptr[2] = val & 0x20? ink: paper;
	ShadowBaseptr[3] = val & 0x10? ink: paper;
	ShadowBaseptr[4] = val & 0x08? ink: paper;
	ShadowBaseptr[5] = val & 0x04? ink: paper;
	ShadowBaseptr[6] = val & 0x02? ink: paper;
	ShadowBaseptr[7] = val & 0x01? ink: paper;
}

void
screen_write_xim_mode_2_2x2(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns16 offset;
	unsigned int ink, paper;
	unsigned long i_pix, p_pix, tmp;
	uns8 attr;

	uns8 pmask, ipmask;
	int j, k;

	screen_dirty++;

	offset = addr;

	y = (offset & 0xffe0) >> 4;
	x = (offset & 0x001f) << 4;
	
	attr = realMem[VMPR_SCREEN_PAGE][8192 + addr];
	ink = Foreground(attr); paper = Background(attr);
	/* toggle the colours if we're in the inverse part of the FLASH cycle */
	if (flash_phase && (attr & 0x80)) {
		tmp=ink; ink=paper; paper=tmp;
	}

	i_pix = pixels[clut[ink]];
	p_pix = pixels[clut[paper]];


	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x + 1;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y + 1;

	ByteOffset = x + y*HSize;
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	for (j = 0; j < 2; j++)
		for (k = 0; k < 2; k++) {
		        pixel_offset = j * HSize + k; 
			baseptr[        pixel_offset] = val & 0x80? i_pix: p_pix;
			baseptr[1 * 2 + pixel_offset] = val & 0x40? i_pix: p_pix;
			baseptr[2 * 2 + pixel_offset] = val & 0x20? i_pix: p_pix;
			baseptr[3 * 2 + pixel_offset] = val & 0x10? i_pix: p_pix;
			baseptr[4 * 2 + pixel_offset] = val & 0x08? i_pix: p_pix;
			baseptr[5 * 2 + pixel_offset] = val & 0x04? i_pix: p_pix;
			baseptr[6 * 2 + pixel_offset] = val & 0x02? i_pix: p_pix;
			baseptr[7 * 2 + pixel_offset] = val & 0x01? i_pix: p_pix;
			ShadowBaseptr[        pixel_offset] = val & 0x80? ink: paper;
			ShadowBaseptr[1 * 2 + pixel_offset] = val & 0x40? ink: paper;
			ShadowBaseptr[2 * 2 + pixel_offset] = val & 0x20? ink: paper;
			ShadowBaseptr[3 * 2 + pixel_offset] = val & 0x10? ink: paper;
			ShadowBaseptr[4 * 2 + pixel_offset] = val & 0x08? ink: paper;
			ShadowBaseptr[5 * 2 + pixel_offset] = val & 0x04? ink: paper;
			ShadowBaseptr[6 * 2 + pixel_offset] = val & 0x02? ink: paper;
			ShadowBaseptr[7 * 2 + pixel_offset] = val & 0x01? ink: paper;
		}
}

void
screen_write_xim_mode_3_1x1(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns8 val_0, val_1, val_2, val_3;
	unsigned long pix_0, pix_1, pix_2, pix_3;

	screen_dirty++;

        y = addr >> 7;
        x = (addr & 0x7f) << 1;

	val_0 = HMPR_MD3COL | MD3_SWITCH((val & 0xc0) >> 6);
	val_1 = HMPR_MD3COL | MD3_SWITCH((val & 0x30) >> 4);
	val_2 = HMPR_MD3COL | MD3_SWITCH((val & 0x0c) >> 2);
	val_3 = HMPR_MD3COL | MD3_SWITCH( val & 0x03);
	pix_0 = pixels[clut[val_0]];
	pix_1 = pixels[clut[val_1]];
	pix_2 = pixels[clut[val_2]];
	pix_3 = pixels[clut[val_3]];
		
	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y;

	ByteOffset = x + y*HSize*2;
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	baseptr[0] = pix_0;
/*	baseptr[0] = pix_1; */
	baseptr[1] = pix_2;
/*	baseptr[1] = pix_3; */
	ShadowBaseptr[0] = val_0;
/*	ShadowBaseptr[0] = val_1; */
	ShadowBaseptr[1] = val_2;
/*	ShadowBaseptr[1] = val_3; */
}

void
screen_write_xim_mode_3_2x2(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns8 val_0, val_1, val_2, val_3;
	unsigned long pix_0, pix_1, pix_2, pix_3;
	int j, k;

	screen_dirty++;

        y = (addr & 0xff80) >> 6;
        x = (addr & 0x007f) << 2;

	val_0 = HMPR_MD3COL | MD3_SWITCH((val & 0xc0) >> 6);
	val_1 = HMPR_MD3COL | MD3_SWITCH((val & 0x30) >> 4);
	val_2 = HMPR_MD3COL | MD3_SWITCH((val & 0x0c) >> 2);
	val_3 = HMPR_MD3COL | MD3_SWITCH( val & 0x03);
	pix_0 = pixels[clut[val_0]];
	pix_1 = pixels[clut[val_1]];
	pix_2 = pixels[clut[val_2]];
	pix_3 = pixels[clut[val_3]];
		

	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x + 1;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y + 1;

	ByteOffset = x + y*HSize;
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	for (j = 0; j < 2; j++) {
		        pixel_offset = j * HSize; 
			baseptr[    pixel_offset    ] = pix_0;
			baseptr[    pixel_offset + 1] = pix_1;
			baseptr[2 + pixel_offset    ] = pix_2;
			baseptr[2 + pixel_offset + 1] = pix_3;
			ShadowBaseptr[    pixel_offset    ] = val_0;
			ShadowBaseptr[    pixel_offset + 1] = val_1;
			ShadowBaseptr[2 + pixel_offset    ] = val_2;
			ShadowBaseptr[2 + pixel_offset + 1] = val_3;
	}
}

void
screen_write_xim_mode_4_1x1(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns8  val_0, val_1;
	unsigned long pix_0, pix_1;

	screen_dirty++;

        y = addr >> 7;
        x = (addr & 0x7f) << 1;

	val_0 = (val & 0xf0) >> 4;
	val_1 =  val & 0x0f;
        pix_0 = pixels[clut[val_0]];
        pix_1 = pixels[clut[val_1]];
                     
	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y;

	ByteOffset = x + y*HSize*2;				
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;

	baseptr[0] = pix_0;
	baseptr[1] = pix_1;
	ShadowBaseptr[0] = val_0;
	ShadowBaseptr[1] = val_1;
}

void
screen_write_xim_mode_4_2x2(uns16 addr, uns8 val)
{
	int x, y, ByteOffset, pixel_offset;
	uns8 *baseptr, *ShadowBaseptr;
	uns8  val_0, val_1;
	unsigned long pix_0, pix_1;
	int j, k;

	screen_dirty++;

        y = (addr & 0xff80) >> 6;
        x = (addr & 0x007f) << 2;

	val_0 = (val & 0xf0) >> 4;
	val_1 =  val & 0x0f;
        pix_0 = pixels[clut[val_0]];
        pix_1 = pixels[clut[val_1]];
                     
	/* Extend the bounding box if necessary */
	if (x < x_min) x_min = x;
	if (x > x_max) x_max = x + 1;
	if (y < y_min) y_min = y;
	if (y > y_max) y_max = y + 1;

	ByteOffset = x + y*HSize;				
	baseptr = (uns8 *)xim_data_ptr + ByteOffset;
	ShadowBaseptr = (uns8 *)shadow_image + ByteOffset;
	for (j = 0; j < 2; j++)
		for (k = 0; k < 2; k++) {
		        pixel_offset = j * HSize + k; 
			baseptr[    pixel_offset] = pix_0;
			baseptr[2 + pixel_offset] = pix_1;
			ShadowBaseptr[    pixel_offset] = val_0;
			ShadowBaseptr[2 + pixel_offset] = val_1;
	}
}


#ifdef MITSHM
/* Screen refresh functions - two different functions */
void 
screen_refresh_xshm(void)
{
	if (!screen_dirty) return;

	x_max |= round_up_x[ScaleFactor - 1];

	XShmPutImage(dpy, mainWin, theGc, xim, x_min, y_min, x_min, y_min,
		x_max - (x_min - 1), y_max - (y_min - 1), False);
	XFlush(dpy);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;
	screen_dirty = 0;
}
#endif

void 
screen_refresh_xim(void)
{
	if (!screen_dirty) return;

	x_max |= round_up_x[ScaleFactor - 1];

	XPutImage(dpy, mainWin, theGc, xim, x_min, y_min, x_min, y_min,
		x_max - (x_min - 1), y_max - (y_min - 1));
	XFlush(dpy);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;
	screen_dirty = 0;
}


void
force_screen_refresh(void)
{
	int i;

        switch (VMPR_SCREEN_MODE) {
        
         case MODE_1:
  	  for (i = 0; i < 768; i++) {
	  	attr_write_mode_1(6144 + i, realMem[VMPR_SCREEN_PAGE][6144 + i]);
	  }
	  break;
         case MODE_2:
  	  for (i = 0; i < 6144; i++) {
	  	attr_write_mode_2(8192 + i, realMem[VMPR_SCREEN_PAGE][8192 + i]);
	  }
	  break;
         case MODE_3:
         case MODE_4:
  	  for (i = 0; i < 0x4000; i++) {
	  	screen_write(i, realMem[VMPR_SCREEN_PAGE][i]);
	  }
  	  for (i = 0; i < 0x2000; i++) {
	  	screen_write(0x4000+i, realMem[(VMPR_SCREEN_PAGE+1) &0x1f][i]);
	  }
          break;
        } /* End Switch */
}


void
set_screen_write_pointers(void)
{
	if (ScaleFactor ==1) {
	
		screen_write_xim_mode_1 = screen_write_xim_mode_1_1x1;
		screen_write_xim_mode_2 = screen_write_xim_mode_2_1x1;
		screen_write_xim_mode_3 = screen_write_xim_mode_3_1x1;
		screen_write_xim_mode_4 = screen_write_xim_mode_4_1x1;

	}
	else {
	
		screen_write_xim_mode_1 = screen_write_xim_mode_1_2x2;
		screen_write_xim_mode_2 = screen_write_xim_mode_2_2x2;
		screen_write_xim_mode_3 = screen_write_xim_mode_3_2x2;
		screen_write_xim_mode_4 = screen_write_xim_mode_4_2x2;
	
	}

}


/* Perform a clean close-down. */
static void
closedown(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		XShmDetach(dpy, &xshminfo);
		XDestroyImage(xim);
		shmdt(xshminfo.shmaddr);
		shmctl(xshminfo.shmid, IPC_RMID, 0);
	}
#endif

/* Free up the Colours */

	XFreeColors(dpy, cmap, pixels, NCOLORS, 0L);

	XAutoRepeatOn(dpy);
	XCloseDisplay(dpy);
}


#ifdef SLOWDOWN
/* Speed up or slow down the emulator */
static void
adjust_speed(int delta)
{
	int newspd = GetCfg(slowdown) + delta;

	if (newspd < 0) newspd = GetCfg(slowdown);

	SetCfg(slowdown, newspd);
}
#endif


/* Process a KeyPress event */
static void
process_keypress(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;
	char *str;
	int i;	


	XLookupString(kev, buf, 2, &ks, NULL);

	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] |= 0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] |= 0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] |= 0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] |= 0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] |= 0x01;
#ifdef SLOWDOWN
	else if (ks == key_table[KM_KEY_SLOWER])
		adjust_speed(1);
	else if (ks == key_table[KM_KEY_FASTER])
		adjust_speed(-1);
#endif

	switch (ks) {
		/* Emulator control keys */
		case XK_F1: 
			InterruptType = Z80_reset;
			for (i = 0; i < 9; i++)
				keyports[i] = 0xff;
			break;
		case XK_F2:	/* Switch Keyboard mode */
			SpectrumKeyboard = !SpectrumKeyboard;
			break;
		case XK_F3:	/* Reset Mouse / pointer */ 
			if (!SAMPointerMode) {
				XDefineCursor(dpy, mainWin, SAMNullCursor);
				SAMButtons  = 0xff;
				XWarpPointer(dpy, mainWin, mainWin, 0, 0, 0, 0,
					     SAMPointerX, SAMPointerY);
			}
			SAMPointerMode = (SAMPointerMode + 1) % 3;
			if (!SAMPointerMode) {
				XUndefineCursor(dpy, mainWin);
			}
			break;

		case XK_F4: /* Turn on/off palette switching */
			DoPaletteSwitching = !DoPaletteSwitching;
			break;
		case XK_F5:	/* Rescale windows */
			ScaleFactor = (ScaleFactor == 1) ? 2 : 1;
			set_screen_write_pointers();

			/* Call vmpr to reset the screen write pointers and
			   fetch the screen from memory */
			out_vmpr(vmpr);

			if (GetCfg(NoBorder)) {
			  XResizeWindow(dpy, borderWin, HSize, VSize);
			  XMoveResizeWindow(dpy, mainWin, 0, 0, HSize, VSize);
			}
			else {
			  XResizeWindow(dpy, borderWin, HSize + BorderWidth*2, 
						        VSize + BorderWidth*2);
			  XMoveResizeWindow(dpy, mainWin, BorderWidth,
					    BorderWidth, HSize, VSize);
			}
			/* force a refresh of the entire screen */
			x_max = HSize - 1; y_max = VSize - 1; x_min = y_min = 0;
			screen_dirty=1;
			screen_refresh();
			break;


		case XK_F6:     /* Insert new floppy disk in drive 1 */
			str = dialog_event_loop("Disk Image for drive 1");
			if (strlen(str))
				insert_disk(str, 1);
			break;
		case XK_F7:     /* Insert new floppy disk in drive 2 */
			str = dialog_event_loop("Disk Image for drive 2");
			if (strlen(str))
				insert_disk(str, 2);
			break;
		case XK_F8:	/* Generate an NMI */
			InterruptType = Z80_nmi;
			break;
			
		/* Crude Debug toggle */
		case XK_F9:
			str = dialog_event_loop("Toggle debug");
			disass = !disass;
			break;
		case XK_F10: case XK_F12:
			Quit(0);
			break;

		/* Ordinary SAM keys */
		case XK_Tab:
			keyports[3] &= 0xbf; break;
		case XK_Return:
			keyports[6] &= 0xfe; break;
		case XK_Control_L: case XK_Control_R:
		/* <Shift> <Control> is mapped to Caps Lock */
			if (keyports[0] & 0x01)
			     keyports[8] &= 0xfe;
			else
			     keyports[3] &= 0x7f;
			break;
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] &= 0xfe; break;
		case XK_Meta_L:
		case XK_Alt_L:
			keyports[7] &= 0xfd; break;
#ifdef DEC
		/* Use DEC "Compose Character" key for <edit> if available
		   else use RHS Alt/Meta key*/
		case XK_Multi_key: 
#endif
		case XK_Alt_R: case XK_Meta_R: case XK_Mode_switch:
			keyports[6] &= 0x7f; break;
		case XK_BackSpace: case XK_Delete:
			keyports[4] &= 0x7f; break;
		case XK_Escape: case XK_grave:
			keyports[3] &= 0xdf; break;
		case XK_Up:
			keyports[8] &= 0xfd; break;
		case XK_Down:
			keyports[8] &= 0xfb; break;
		case XK_Left:
			keyports[8] &= 0xf7; break;
		case XK_Right:
			keyports[8] &= 0xef; break;
		case XK_minus:
			keyports[4] &= 0xdf; break;
		case XK_underscore:
			keyports[5] &= 0xdf; break;
		case XK_equal:
			keyports[5] &= 0xdf; break;
		case XK_plus:
			keyports[0] |= 0x01; 
			keyports[4] &= 0xbf; break;
		case XK_semicolon:
			keyports[6] &= 0xdf; break;
		case XK_colon:
			keyports[0] |= 0x01; 
			keyports[6] &= 0xbf; break;
		case XK_apostrophe:
			keyports[0] &= 0xfe; 
			keyports[4] &= 0xf7; break;
		case XK_comma:
			keyports[7] &= 0xdf; break;
		case XK_period:
			keyports[7] &= 0xbf; break;
		case XK_less:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xfe; break;
		case XK_greater:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xfd; break;
		case XK_slash:
			keyports[0] &= 0xfe; keyports[4] &= 0xdf; break;
		case XK_question:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[0] &= 0xfb; break;
		case XK_backslash:
		        keyports[0] &= 0xfe; keyports[7] &= 0x7f;
			break;
		case XK_bar:
		        keyports[0] |= 0x01; keyports[7] &= 0xfd;
		        keyports[4] &= 0xfd;
			break;
                case XK_ampersand:
                        keyports[4] &=0xef; break;
                case XK_bracketleft:
                	keyports[7] &= 0xfd; keyports[2] &= 0xf7;
			break;
                case XK_bracketright:
                	keyports[7] &= 0xfd; keyports[2] &= 0xef;
			break;
                case XK_braceleft:
                        keyports[0] |= 0x01;
                        keyports[7] &= 0xfd; keyports[1] &= 0xf7;
                        break;                        
                case XK_braceright:
                        keyports[0] |= 0x01;
                        keyports[7] &= 0xfd; keyports[1] &= 0xef;
                        break;                        
		case XK_quotedbl:
			keyports[0] |= 0x01;
			keyports[5] &= 0xbf;
			break;
		case XK_asciitilde:
			keyports[4] &= 0xfe;
			break;
		case XK_asciicircum:
		        keyports[0] |= 0x01; keyports[7] &= 0xfd;
		        keyports[6] &= 0xef;
		        break;
		/* hash and sterling characters must be treated 
		   separately as their position depends on the 
		   type of keyboard - UK or US */
		case XK_sterling:
                        keyports[0] |= 0x01;
                        keyports[7] &= 0xfd; keyports[6] &= 0xfd;
			break;
		case XK_numbersign:
                        keyports[0] |= 0x01;
			keyports[0] &= 0xfe; keyports[3] &=0xfb;
			break;

		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case XK_KP_0:
			keyports[5] &= 0x7f;
			break;
		case XK_KP_1:
			keyports[0] &= 0xdf;
			break;
		case XK_KP_2:
			keyports[0] &= 0xbf;
			break;
		case XK_KP_3:
			keyports[0] &= 0x7f;
			break;
		case XK_KP_4:
			keyports[1] &= 0xdf;
			break;
		case XK_KP_5:
			keyports[1] &= 0xbf;
			break;
		case XK_KP_6:
			keyports[1] &= 0x7f;
			break;
		case XK_KP_7:
			keyports[2] &= 0xdf;
			break;
		case XK_KP_8:
			keyports[2] &= 0xbf;
			break;
		case XK_KP_9:
			keyports[2] &= 0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case XK_KP_Enter:
                        keyports[7] &= 0x7f;
                	break;
                case XK_KP_Decimal:
			keyports[0] &= 0xfe; keyports[5] &= 0xbf;
                	break;

		default:
			if (ks < 128) {
				keyports[keytable[ks].port] &= keytable[ks].bit;
			}
			break;
	}
}

/* Process a KeyRelease event */
static void
process_keyrelease(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] &= ~0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] &= ~0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] &= ~0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] &= ~0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] &= ~0x01;

	switch (ks) {

		/* Ordinary SAM keys */
		case XK_Tab:
			keyports[3] |= ~0xbf; break;

		case XK_Return:
			keyports[6] |= ~0xfe; break;
		case XK_Control_L: case XK_Control_R:
		/* <Shift> <Control> is mapped to Caps Lock */
			     keyports[8] |= ~0xfe;
			     keyports[3] |= ~0x7f;
			break;
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] |= ~0xfe; break;
		case XK_Meta_L:
		case XK_Alt_L: 
			keyports[7] |= ~0xfd; break;
#ifdef DEC
		/* Use DEC "Compose Character" key for <edit> if available
		   else use RHS Alt/Meta key*/
		case XK_Multi_key: 
#endif
		case XK_Alt_R: case XK_Meta_R: case XK_Mode_switch:
			keyports[6] |= ~0x7f; break;
		case XK_BackSpace: case XK_Delete:
			keyports[4] |= ~0x7f; break;
		case XK_Escape: case XK_grave:
			keyports[3] |= ~0xdf; break;
		case XK_Up:
			keyports[8] |= ~0xfd; break;
		case XK_Down:
			keyports[8] |= ~0xfb; break;
		case XK_Left:
			keyports[8] |= ~0xf7; break;
		case XK_Right:
			keyports[8] |= ~0xef; break;
		case XK_minus:
			keyports[4] |= ~0xdf; break;
		case XK_underscore:
			keyports[5] |= ~0xdf; break;
		case XK_equal:
			keyports[5] |= ~0xdf; break;
		case XK_plus:
			keyports[4] |= ~0xbf; break;
		case XK_semicolon:
			keyports[6] |= ~0xdf; break;
		case XK_colon:
			keyports[6] |= ~0xbf; break;
		case XK_apostrophe:
			keyports[0] |= ~0xfe; 
			keyports[4] |= ~0xf7; break;
		case XK_comma:
			keyports[7] |= ~0xdf; break;
		case XK_period:
			keyports[7] |= ~0xbf; break;
		case XK_less:
			keyports[7] |= ~0xfd;
			keyports[2] |= ~0xfe; break;
		case XK_greater:
			keyports[7] |= ~0xfd;
			keyports[2] |= ~0xfd; break;
		case XK_slash:
			keyports[0] |= ~0xfe; 
			keyports[4] |= ~0xdf; break;
		case XK_question:
			keyports[7] |= ~0xfd;
			keyports[0] |= ~0xfb; break;
		case XK_backslash:
		        keyports[0] |= ~0xfe; keyports[7] |= ~0x7f;
			break;
		case XK_bar:
		        keyports[7] |= ~0xfd;
		        keyports[4] |= ~0xfd;
			break;
                case XK_ampersand:
                        keyports[4] |= ~0xef; break;
                case XK_bracketleft:
                	keyports[7] |= ~0xfd; keyports[2] |= ~0xf7;
			break;
                case XK_bracketright:
                	keyports[7] |= ~0xfd; keyports[2] |= ~0xef;
			break;
                case XK_braceleft:
                        keyports[7] |= ~0xfd; keyports[1] |= ~0xf7;
                        break;                        
                case XK_braceright:
                        keyports[7] |= ~0xfd; keyports[1] |= ~0xef;
                        break;                        
		case XK_quotedbl:
			keyports[5] |= ~0xbf;
			break;
		case XK_asciitilde:
			keyports[4] |= ~0xfe;
			break;
		case XK_asciicircum:
		        keyports[7] |= ~0xfd;
		        keyports[6] |= ~0xef;
		        break;
		/* hash and sterling characters must be treated 
		   separately as their position depends on the 
		   type of keyboard - UK or US */
		case XK_sterling:
                        keyports[7] |= ~0xfd; keyports[6] |= ~0xfd;
			break;
		case XK_numbersign:
			keyports[0] |= ~0xfe; keyports[3] |= ~0xfb;
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case XK_KP_0:
			keyports[5] |= ~0x7f;
			break;
		case XK_KP_1:
			keyports[0] |= ~0xdf;
			break;
		case XK_KP_2:
			keyports[0] |= ~0xbf;
			break;
		case XK_KP_3:
			keyports[0] |= ~0x7f;
			break;
		case XK_KP_4:
			keyports[1] |= ~0xdf;
			break;
		case XK_KP_5:
			keyports[1] |= ~0xbf;
			break;
		case XK_KP_6:
			keyports[1] |= ~0x7f;
			break;
		case XK_KP_7:
			keyports[2] |= ~0xdf;
			break;
		case XK_KP_8:
			keyports[2] |= ~0xbf;
			break;
		case XK_KP_9:
			keyports[2] |= ~0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case XK_KP_Enter:
                        keyports[7] |= ~0x7f;
                	break;
                case XK_KP_Decimal:
			keyports[0] |= ~0xfe; keyports[5] |= ~0xbf;
                	break;

		default:
			if (ks < 128)
				keyports[keytable[ks].port] |= ~keytable[ks].bit;
			break;
	}
}

/* Process a pointer MotionNotify event */
static void
process_pointermove(XPointerMovedEvent *pev)
{
	int DeltaX, DeltaY;

	DeltaX = ((pev->x - SAMPointerX) * SAMPointerMode) / ScaleFactor;
	DeltaY = (SAMPointerY - pev->y) / ScaleFactor;

	mouse_xy_write(DeltaX, DeltaY);	
	XWarpPointer(dpy, mainWin, mainWin, 0, 0, 0, 0, SAMPointerX, SAMPointerY);

}

/* Process mouse ButtonPressed event */
static void
process_buttonpress(XButtonPressedEvent *bev)
{
	switch (bev->button) {

		case Button1:
			SAMButtons &= ~SAM_BUTTON_1;
			break;				
		case Button2:
			SAMButtons &= ~SAM_BUTTON_2;
			break;				
		case Button3:
			SAMButtons &= ~SAM_BUTTON_3;
			break;
		case Button4:
			SAMButtons &= ~SAM_BUTTON_4;
			break;				

	} /* end switch */
	mouse_button_write(SAMButtons);
}

/* Process mouse ButtonReleased event */
static void
process_buttonrelease(XButtonPressedEvent *bev)
{
	switch (bev->button) {

		case Button1:
			SAMButtons |= SAM_BUTTON_1;
			break;				
		case Button2:
			SAMButtons |= SAM_BUTTON_2;
			break;				
		case Button3:
			SAMButtons |= SAM_BUTTON_3;
			break;
		case Button4:
			SAMButtons |= SAM_BUTTON_4;
			break;				
	} /* end switch */
	mouse_button_write(SAMButtons);
}

/* Tell the window manager everything it needs to know about the 
 * top-level window (borderWin).
 */
static void
notify_wm(int *argcnt, char **argvec)
{
	Pixmap icon_pix;
#ifdef XPM_ICON
	Pixmap shape_pix;
	XpmAttributes icon_xpm_attrs;
#endif
	XWMHints xwmh;
	XSizeHints xsh;
	XClassHint xch;
	XTextProperty appName, iconName;
	char *app_name;
	char *icon_name="xcoupe";

#ifdef XPM_ICON

	icon_xpm_attrs.visual    = DefaultVisual(dpy, scrn);
	icon_xpm_attrs.colormap  = DefaultColormap(dpy,scrn);
	icon_xpm_attrs.valuemask = XpmCloseness | XpmColormap | XpmVisual;
	XpmCreatePixmapFromData(dpy, root, xcoupe_icon_xpm, 
				&icon_pix, &shape_pix, 
				&icon_xpm_attrs);
#else
	icon_pix = XCreatePixmapFromBitmapData(dpy,
		root, xcoupe_bits, xcoupe_width, xcoupe_height,
		BlackPixel(dpy, scrn), WhitePixel(dpy, scrn),
		DefaultDepth(dpy, scrn));

#endif
	app_name = (char *)malloc(20);
	if (!app_name)
		app_name = "xcoupe";
	else
		sprintf(app_name, "xcoupe %d.%d", XCOUPE_Major, XCOUPE_Minor);

	xsh.flags = PSize | PMinSize | PMaxSize;

	if (strlen(GetCfg(geometry)) > 0)
		xsh.flags |= USPosition;
	else
		xsh.flags |= PPosition;

	xsh.min_width  = HSize; xsh.min_height = VSize;
	xsh.max_width  = HSize + BorderWidth * 2;
	xsh.max_height = VSize + BorderWidth * 2;

	if (XStringListToTextProperty(&app_name, 1, &appName) == 0) {
		xcoupe_mesg(XCOUPE_ERR, "can't create a TextProperty!");
		return;
	}
	if (XStringListToTextProperty(&icon_name, 1, &iconName) == 0) {
		xcoupe_mesg(XCOUPE_ERR, "can't create a TextProperty!");
		return;
	}

	xwmh.initial_state = NormalState;
	xwmh.input = True;
	xwmh.icon_pixmap = icon_pix;
	xwmh.flags = StateHint | IconPixmapHint | InputHint;
#ifdef XPM_ICON
	xwmh.icon_mask = shape_pix;
	xwmh.flags |= IconMaskHint;
#endif
	

	xch.res_name = "xcoupe";
	xch.res_class = "Xcoupe";

	XSetWMProperties(dpy, borderWin, &appName, &iconName, 
		argvec, *argcnt, &xsh, &xwmh, &xch);
}


/* Calculate initial window position depending on our geometry spec.
 * FIXME: this does not take into account the size of WM decorations.
 */
static void
calc_win_position(int *x_ret, int *y_ret)
{
	int x_tmp, y_tmp, w_ret, h_ret;
	int mask;

	if (!GetCfg(geometry)[0]) return;

	mask = XParseGeometry(GetCfg(geometry), &x_tmp, &y_tmp, &w_ret, &h_ret);

	*x_ret = 0; *y_ret = 0;

	if (!mask) {
		xcoupe_mesg(XCOUPE_WARN, "invalid geometry specification %s", 
			GetCfg(geometry));
		GetCfg(geometry)[0] = '\0';
		return;
	}

	if (mask & XValue) 
		*x_ret = (mask & XNegative) ?
			(DisplayWidth(dpy, scrn) +  x_tmp - (HSize + BorderWidth* 2)) :
			x_tmp;

	if (mask & YValue) 
		*y_ret = (mask & YNegative) ?
			(DisplayHeight(dpy, scrn) + y_tmp - (VSize + BorderWidth * 2)) :
			y_tmp;
}


/*
 * Image initialisation routines
 */

static int
image_init(void)
{
#ifdef MITSHM
	if (GetCfg(use_mitshm)) {
		xim = XShmCreateImage(dpy, DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, NULL, &xshminfo, HSize, VSize);
		if (!xim) {
			xcoupe_mesg(XCOUPE_ERR, "xcoupe: failed to create a shared XImage!");
			return -1;
		}
		xshminfo.shmid = shmget(IPC_PRIVATE, xim->bytes_per_line * xim->height,
			IPC_CREAT|0777);
		if (xshminfo.shmid == -1) {
			xcoupe_mesg(XCOUPE_ERR, "shmget <xshminfo.shmid> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.shmaddr = xim->data = shmat(xshminfo.shmid, 0, 0);
		if (!xshminfo.shmaddr) {
			xcoupe_mesg(XCOUPE_ERR, "shmat <xshminfo.shmaddr> failed");
			perror("  reason:");
			return -1;
		}
		xshminfo.readOnly = False;
		if (!XShmAttach(dpy, &xshminfo)) {
			xcoupe_mesg(XCOUPE_ERR, "XShmAttach failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh = screen_refresh_xshm;
	} else
#endif
	{
		xim = XCreateImage(dpy,  DefaultVisual(dpy, scrn),
			DefaultDepth(dpy, scrn), ZPixmap, 0, NULL, HSize, VSize, 8, 0);
		if (!xim) {
			xcoupe_mesg(XCOUPE_ERR, "XCreateImage failed");
			perror("  reason:");
			return -1;
		}
		if ((xim->data = (char *)malloc(xim->bytes_per_line * xim->height))
			== NULL) {
			xcoupe_mesg(XCOUPE_ERR, "malloc <xim->data> failed");
			perror("  reason:");
			return -1;
		}
		screen_refresh = screen_refresh_xim;
	}
	screen_write = screen_write_xim_mode_1;
	bytes_per_line = xim->bytes_per_line;
	xim_total_size = xim->bytes_per_line * xim->height;
	xim_data_ptr = xim->data;

	/* Create Shadow 'image' used for fast clut update */
	if ((shadow_image = (char *)malloc(xim->bytes_per_line * xim->height))
		== NULL) {
		xcoupe_mesg(XCOUPE_ERR, "malloc <shadow_image> failed");
		perror("  reason:");
		return -1;
	}

	return 0;
}


/*
 * Initialise the 128 colours required by the SAM
 * Return -1 if colour allocation failed.
 */
static int
get_colours(unsigned long *pix_ret)
{
	XColor xc;
	int i;
        int red_index, green_index, blue_index;

        int RGBvals[8] = {0x0000,0x0c30, 0x30c3, 0x3cf3, 
        	          0xc30c,0xcf3c, 0xf3cf, 0xffff}; 

	xc.flags = DoRed | DoGreen | DoBlue;
	for (i = 0; i < NCOLORS; i++) {

		blue_index  = (i & 0x01) << 1| ((i & 0x10) >> 2) | ((i & 0x08) >> 3);                       
		red_index   = (i & 0x02)     | ((i & 0x20) >> 3) | ((i & 0x08) >> 3);                       
		green_index = (i & 0x04) >> 1| ((i & 0x40) >> 4) | ((i & 0x08) >> 3);                       

	 	xc.red   = RGBvals[red_index];
 		xc.green = RGBvals[green_index];
 		xc.blue  = RGBvals[blue_index];

		if (!XAllocColor(dpy, cmap, &xc)) {
			xcoupe_mesg(XCOUPE_WARN, "XAllocColor failed on colour %d", i);
			return(-1);
		}
		pix_ret[i] = xc.pixel;
	}
	return 0;
}

Cursor CreateNullCursor(Window win)
{
	static char NullCursor_bits[] = { 0x00 };
	Cursor NullCursor;
	Pixmap NullPixmap;
	XColor NullColour;
	
	NullPixmap = XCreatePixmapFromBitmapData(dpy,
		win, NullCursor_bits, 1, 1,
		BlackPixel(dpy, scrn), WhitePixel(dpy, scrn),
		1);

	NullColour.pixel = BlackPixel(dpy,scrn);
	XQueryColor(dpy, DefaultColormap(dpy,scrn), &NullColour);

	NullCursor = XCreatePixmapCursor(dpy, NullPixmap, NullPixmap, 
		     &NullColour, &NullColour, 0, 0);

/*	NullCursor = XCreateFontCursor(dpy,128);*/
	return(NullCursor);
}

/**************************************************************************
 *  Public functions start here
 */

/* Start all the "hardware" up */
void
start_emulator(int *argcp, char **argvp)
{
	int i;

	key_table = (KeySym *)malloc(sizeof(KeySym) * NUM_KEYMAPPINGS);

	x_max = y_max = 0; x_min = HSize - 1; y_min = VSize - 1;

	/* Connect to a display and initialise run-time configuration */
	if ((dpy = xcoupe_open_display(argcp, argvp, key_table)) == NULL) {
		xcoupe_mesg(XCOUPE_FATAL, "failed to open X display");
	}

	flash_phase = !GetCfg(flashing);

	/* Some useful X variables */
	scrn = DefaultScreen(dpy);
	sptr = DefaultScreenOfDisplay(dpy);
	root = DefaultRootWindow(dpy);

	/* Before we go anywhere we had better check the depth of the
	   visual. At the moment XCoupe only allows 8 bits per pixel
	   visuals */
	if (DefaultDepth(dpy, scrn) != 0x08) {
		xcoupe_mesg(XCOUPE_FATAL, "XCoupe only supports visuals with 8 bits per pixel (8 bbp)");
	}

	theGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, theGc);
	XSetGraphicsExposures(dpy, theGc, False);

	fgGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, fgGc);
	XSetGraphicsExposures(dpy, fgGc, False);

	bgGc = XCreateGC(dpy, root, 0, NULL);
	XCopyGC(dpy, DefaultGC(dpy, scrn), ~0, bgGc);
	XSetGraphicsExposures(dpy, bgGc, False);

	/* Colour initialisation */
	if (GetCfg(priv_cmap))
		cmap = XCreateColormap(dpy, root, DefaultVisual(dpy, scrn), AllocNone);
	else
		cmap  = DefaultColormap(dpy, scrn);

	if (get_colours(pixels) == -1) {
		xcoupe_mesg(XCOUPE_WARN, "colour allocation failed - using private cmap");
		SetCfg(priv_cmap, True);
		cmap = XCreateColormap(dpy, root, DefaultVisual(dpy, scrn), AllocNone);
		if (get_colours(pixels) == -1) 
			xcoupe_mesg(XCOUPE_FATAL, "private colourmap initialisation failed!");
	}

	/* Initialise variables used to keep track of line interrupt 
	   clut changes */

	for (i=0; i < 16; i++)
		first_line[i] = 0;
	line_int_disabled = 0;
	clut_change_in_frame = 0;
	border_change_in_frame = 0;
	vmpr_change_in_frame = 0;	
	DoPaletteSwitching = 1;

	/* Set the dynamic scaling to 2x2 initially */

	ScaleFactor = 2;
	set_screen_write_pointers();

	/* Get a font to use */
	theFont = XLoadQueryFont(dpy, GetCfg(font));
	if (!theFont) {
		xcoupe_mesg(XCOUPE_WARN, "can't load font <%s>, trying <%s>...",
			GetCfg(font), DEF_FONT);
		theFont = XLoadQueryFont(dpy, DEF_FONT);
		if (!theFont)
			xcoupe_mesg(XCOUPE_FATAL, "can't load font <%s>!", DEF_FONT);
	}
	XSetFont(dpy, theGc, theFont->fid);

	/* Image handling initialisation */
	if (image_init() != 0) {
#ifdef MITSHM
		xcoupe_mesg(XCOUPE_WARN,
				"MIT-SHM initialisation failure, trying without MIT-SHM...");
		SetCfg(use_mitshm, False);
#endif
		if (image_init() != 0)
			xcoupe_mesg(XCOUPE_FATAL, "image initialisation failure!");
	}

	/* No keys initially being pressed */
	for (i = 0; i < 9; i++)
		keyports[i] = 0xff;
	/* Keyboard starts off in SAM mode by default */
	SpectrumKeyboard = 0;

#if defined(PCSPKR_AUDIO) || defined(DEV_AUDIO) || defined(NAS_AUDIO)
	if (audio_init() != 0) {
		xcoupe_mesg(XCOUPE_WARN, "audio initialisation failed");
		sound_on = -1;
	} else {
		sound_on = GetCfg(use_sound);
	}
#endif

#ifdef JOY
	/* Linux analogue joystick initialisation */
	if (GetCfg(use_js) && joystick_init() != 0) { 
		xcoupe_mesg(XCOUPE_WARN, "joystick initialisation failed");
		SetCfg(use_js, 0);
	}
#endif

	/* Mouse initialisation */
	mouse_init();


	calc_win_position(&xpos, &ypos);

	/* Window creation : if the NoBorder option has been used then the
	   border window is hidden under the main window */

	if (GetCfg(NoBorder)) {
		borderWin = XCreateSimpleWindow(dpy, root, xpos, ypos,
			HSize, VSize,
			0, pixels[0], pixels[0]);
		mainWin = XCreateSimpleWindow(dpy, borderWin, 0, 0,
			HSize, VSize, 0, pixels[0], pixels[0]); /* black */
	}
	else {
		borderWin = XCreateSimpleWindow(dpy, root, xpos, ypos,
			HSize + BorderWidth * 2, VSize + BorderWidth * 2 ,
			0, pixels[0], pixels[0]);
		mainWin = XCreateSimpleWindow(dpy, borderWin, BorderWidth, BorderWidth,
			HSize, VSize, 0, pixels[0], pixels[0]); /* black */
	}
	
	make_dialog(argcp, argvp);

	/* let's be environmentally friendly */
	notify_wm(argcp, argvp);

	if (GetCfg(priv_cmap))
		XSetWindowColormap(dpy, borderWin, cmap);
	
	XSelectInput(dpy, borderWin,  ExposureMask |
		 StructureNotifyMask);
	XSelectInput(dpy, mainWin, ExposureMask | KeyPressMask | KeyReleaseMask | 
		PointerMotionMask | ButtonPressMask | 
		ButtonReleaseMask | EnterWindowMask |  LeaveWindowMask);
	XMapRaised(dpy, borderWin);
	XMapRaised(dpy, mainWin);

	/* Create a blank X cursor for when the SAM mouse is active */
	SAMNullCursor = CreateNullCursor(mainWin);

	XFlush(dpy);

	OnQuit(closedown);
}


/* Check & process any incoming X events */
void 
check_events(void)
{
	static XEvent xev;
	XConfigureEvent *conf_ev;
	XCrossingEvent *cev;
	XPointerMovedEvent *pev, *PointerEvent;

	PointerEvent = NULL;

	while (XEventsQueued(dpy, QueuedAfterFlush)) {
		XNextEvent(dpy, &xev);
		switch(xev.type) {
			case Expose:
				/* force a refresh of the entire screen */
				x_max = HSize - 1; y_max = VSize - 1; x_min = y_min = 0;
				screen_dirty=1;
				screen_refresh();
				break;
			case ConfigureNotify:
				conf_ev = (XConfigureEvent *)&xev;
				XMoveWindow(dpy, mainWin, (conf_ev->width - HSize) / 2,
				(conf_ev->height - VSize) / 2);
				break;
			case MapNotify: case UnmapNotify: case ReparentNotify:
				/* just avoid complaints about unhandled events here */
				break;
			case EnterNotify: case LeaveNotify:
				cev = (XCrossingEvent *)&xev;
				if (cev->type == LeaveNotify) 
					XAutoRepeatOn(dpy);
				else 
					XAutoRepeatOff(dpy);
				break;
			case KeyPress:
				if (!SpectrumKeyboard)
					process_keypress((XKeyEvent *)&xev);
				else	
					process_spectrum_keypress((XKeyEvent *)&xev);
				break;
			case KeyRelease:
				if (!SpectrumKeyboard)
					process_keyrelease((XKeyEvent *)&xev);
				else	
					process_spectrum_keyrelease((XKeyEvent *)&xev);
				break;
			case MotionNotify:
				pev = (XPointerMovedEvent *)&xev;
				if (pev->window == mainWin) {
					PointerEvent = pev;
				}
				break;
			case ButtonPress:
				process_buttonpress((XButtonPressedEvent *)&xev);
				break;
			case ButtonRelease:
				process_buttonrelease((XButtonReleasedEvent *)&xev);
				break;
			default:
				xcoupe_mesg(XCOUPE_WARN, "unhandled X event, type %d", xev.type);
				break;
		}
	}

	/* We only process the last MotionNotify event - so the SAM mouse
	   code can keep up! */
	if (PointerEvent != NULL && SAMPointerMode) {
		process_pointermove(PointerEvent); 
		PointerEvent = NULL;
	}
	
}


/* <val> has just been written to <addr> in the attribute area.
 * Update the screen with screen_write(). Different routines needed for
 * modes 1 and 2.
 */
void
attr_write_mode_1(uns16 addr, uns8 val)
{
	int a_off = (addr & 0x3fff) - 6144;
	int d_off;

	d_off = ((a_off & 0x300) << 3) +    /* Which third of the display file */
			(a_off & 0xff);

	screen_write(d_off + 0x000, realMem[VMPR_SCREEN_PAGE][d_off+0x000]);
	screen_write(d_off + 0x100, realMem[VMPR_SCREEN_PAGE][d_off+0x100]);
	screen_write(d_off + 0x200, realMem[VMPR_SCREEN_PAGE][d_off+0x200]);
	screen_write(d_off + 0x300, realMem[VMPR_SCREEN_PAGE][d_off+0x300]);
	screen_write(d_off + 0x400, realMem[VMPR_SCREEN_PAGE][d_off+0x400]);
	screen_write(d_off + 0x500, realMem[VMPR_SCREEN_PAGE][d_off+0x500]);
	screen_write(d_off + 0x600, realMem[VMPR_SCREEN_PAGE][d_off+0x600]);
	screen_write(d_off + 0x700, realMem[VMPR_SCREEN_PAGE][d_off+0x700]);

}
void
attr_write_mode_2(uns16 addr, uns8 val)
{
	int d_off = (addr & 0x3fff) - 8192;

	screen_write(d_off, realMem[VMPR_SCREEN_PAGE][d_off]);

}


/*
 * Input & output port emulation.
 */

void 
init_ports(void) 
{
	int i;

	vmpr = 0x81;
	hmpr = 0x01;
	lmpr = 0x1f;
	midi = 0x00;	
	border = 0x00;

	for (i=0; i < 16; i++)
		clut[i] = i | (i & 0x07)<<4;
		
	keyboard = 0x5f;
	status_reg = 0xff;
	
	line_int = 0xff;
        hpen = 0x00;
        lpen = 0x00;
        	
        clock_reg = 0x00;

	lepr = 0x00;
	hepr = 0x00;
	attr_reg = 0xff;

}

static inline void 
out_hepr(uns8 val)
{
	int ExternalPage;

        hepr = val;

/* If somebody tries to access non existent external memory, page in the 
   ROM in sections C or D. This means that read-write tests on the
   memory will fail - as required
*/
	
	if (HMPR_MCNTRL) {
     
		if ((ExternalPage = NMAINPAGES+val) >= (ROM0-(GetCfg(externalMem)<<6)))
		   page_in(3, ExternalPage);
		else
	   	   page_in(3, ROM0);
	}
}

static inline void 
out_lepr(uns8 val)
{

	int ExternalPage;

        lepr = val;


/* If somebody tries to access non existent external memory, page in the 
   ROM in sections C or D. This means that read-write tests on the
   memory will fail - as required
*/
        
       	if (HMPR_MCNTRL) {

		if ((ExternalPage = NMAINPAGES+val) >= (ROM0-(GetCfg(externalMem)<<6)))
		   page_in(2, ExternalPage);
		else
		   page_in(2, ROM0);
	}
}

static inline void 
out_vmpr(uns8 val)
{
        vmpr = (val & 0x7f) | 0x80 ;

	switch (VMPR_SCREEN_MODE) {
	
	 case MODE_1:
	  screen_write = screen_write_xim_mode_1;
	  break;
	 case MODE_2:
	  screen_write = screen_write_xim_mode_2;
	  break;
	 case MODE_3:
	  screen_write = screen_write_xim_mode_3;
	  break;
	 case MODE_4:
	  screen_write = screen_write_xim_mode_4;
	  break;

	} /* End switch */
	if (!BORD_SOFF)
	 force_screen_refresh();
}

static inline void 
out_hmpr(uns8 val)
{
        hmpr = val;
        
	if (LMPR_ROM1) 
	   page_in(3,ROM1);
	else
	   page_in(3,(HMPR_PAGE + 1)& 0x1f);
  
	page_in(2, HMPR_PAGE);

   
	if (HMPR_MCNTRL) {
		out_hepr(hepr);
		out_lepr(lepr);
	}

}


static inline void 
out_lmpr(uns8 val)
{
        lmpr = val;
        
	if (LMPR_RAM0)
	   page_in(0, LMPR_PAGE);
	else
	   page_in(0, ROM0);

        page_in(1, (LMPR_PAGE + 1) & 0x1f);

        if (LMPR_ROM1)
           page_in(3, ROM1);
        else
           page_in(3, (HMPR_PAGE + 1) &0x1f);
           
}

/* When line interupts are active, On flyback we need to update the display
   down to the first line_interupt line
*/

void 
update_to_first_line(void)
{

	register int icol, j,i = 0;
	int not_done = 0;
	int FirstBorderLine = 0;
	
	if (line_int_disabled && line_int_disabled++ > 1) {
		
		for (i = 0; i < xim_total_size; i++) {
			*((uns8 *)xim_data_ptr + i) = pixels[clut[*((uns8 *)shadow_image + i)]];
		}
		line_int_disabled = 0;
		y_min = x_min = 0;
		y_max = VSize - 1; x_max = HSize - 1;
		if (!GetCfg(NoBorder)) {
			XSetWindowBackground(dpy, borderWin, pixels[clut[BORD_VAL(border)]]);
			XClearWindow(dpy, borderWin);
		}
		screen_dirty = 1;
	}
	else {

 	 do {
		for (j = 0; j < bytes_per_line*ScaleFactor; j++) {
		    for (icol = 0; icol < 16; icol++) {
			if (i < first_line[icol]-30) {
				not_done = 1;
				if (*((uns8 *)shadow_image + i*bytes_per_line*ScaleFactor + j) == icol)
					*((uns8 *)xim_data_ptr + i*bytes_per_line*ScaleFactor + j) = pixels[clut[icol]];

			}
		    }
		}
	 } while (not_done && ++i < 192);
	 if (!GetCfg(NoBorder)) {
		XSetWindowBackground(dpy, borderWin, pixels[clut[BORD_VAL(border)]]);
		if (FirstBorderLine = first_line[BORD_VAL(border)]*ScaleFactor)
			XClearArea(dpy, borderWin,0, 0, 0, FirstBorderLine, False);
	 }

	 for (icol = 0; icol < 16; icol++)
		first_line[icol] = 0;	
	}

}

inline void 
update_line(void)
{
	int i;

	for (i=bytes_per_line*(LineNo-30)*ScaleFactor;i < bytes_per_line*(LineNo-29)*ScaleFactor; i++)
	 	*((uns8 *)xim_data_ptr + i) = pixels[clut[*((uns8 *)shadow_image + i)]];

	 /* Update the border. NOTE if we are on the last screen line we need
	    to fill in the rest of the border window */
	 if (!GetCfg(NoBorder)) {
		XSetWindowBackground(dpy, borderWin, pixels[clut[BORD_VAL(border)]]);
		XClearArea(dpy, borderWin, 0, LineNo*ScaleFactor, 0, 
				ScaleFactor*((LineNo < 221) ? 1:31), False);
	}
}

static inline void 
out_clut(uns16 port, uns8 val)
{

	register int i;
	int LineStart;
	uns8 old_val = clut[port] & 0x7f;
  
	if (old_val != (val & 0x7f)) {
	 clut[port] = val & 0x7f;
	 if (!BORD_SOFF) {

	        if (!first_line[port])
		    	first_line[port] = LineNo;

		clut_change_in_frame = 1;
		if (DoPaletteSwitching) {
			y_min = x_min = 0;
			y_max = VSize - 1; x_max = HSize - 1;
			screen_dirty = 1;
		}

	        if (!GetCfg(NoBorder) && (BORD_VAL(border) == port) && (LineNo < 30)) {
			XSetWindowBackground(dpy, borderWin, pixels[clut[BORD_VAL(border)]]);
			XClearArea(dpy, borderWin, 0, LineNo*ScaleFactor, 
			           		   0,(30-LineNo)*ScaleFactor, False);
	        }

	 }
	}
}

void
update_border_to_first_line(void)
{
	XSetWindowBackground(dpy, borderWin, pixels[clut[BORD_VAL(border)]]);
	XClearArea(dpy, borderWin, 0, 0, 0, (border_change_in_frame+1)*ScaleFactor, False);	
}

void
out_byte(uns16 port, uns8 val)
{
/*	printf("out_byte(%d,%d)\n",port,val); */
	/* border colour and (possibly) speaker control */
	switch (port & 0xff) {
	    case BORDER_PORT:
		if (BORD_VAL(border) != BORD_VAL(val)) {
			XSetWindowBackground(dpy, borderWin, pixels[clut[BORD_VAL(val)]]);
			XClearArea(dpy, borderWin, 0, (1+LineNo)*ScaleFactor, 0, 0, False);
			border_change_in_frame = LineNo;
		}
		if (BORD_SOFF && !(val  & 0x80)) 
			force_screen_refresh();
		
		border = val;
		keyboard = (keyboard & 0x7f) |BORD_SOFF;
		break;
	    
	/* banked memory management */

            case VMPR_PORT:
		if ((vmpr & 0x7f) != (val & 0x7f)) out_vmpr(val);
        	break;
            case HMPR_PORT:
		if (hmpr != val) out_hmpr(val);
        	break;
            case LMPR_PORT:
        	if (lmpr != val) out_lmpr(val);
        	break;
	    case CLUT_BASE_PORT:
       		out_clut(port >> 8, val);
       		break;
        	
	/* External memory */

	    case HEPR_PORT:
        	out_hepr(val);
        	break;
	    case LEPR_PORT:
        	out_lepr(val);
        	break;
 	
/* Make sure line interupts which are disabled correcly when the line
   is set to > 191
*/  
	    case LINE_PORT:
        	if (val > 191 && line_int < 192)
        		line_int_disabled = 1;
        	if (val < 192 && line_int > 192)
        		line_int_disabled = 0;
		line_int = val;
		break;
	} /* End switch */

/* Disk IO
*/
        if (((port & 0xff) >= DISK1_PORT &&
             (port & 0xff) < DISK1_PORT + 8) ||
            ((port & 0xff) >= DISK2_PORT &&
             (port & 0xff) < DISK2_PORT + 8)) 
             	fdi_out(port,val);

#ifdef PSEUDO_IO
	if (port == PSEUDO_STDOUT)
		putc((int)val, stdout);
	else if (port == PSEUDO_STDERR)
		putc((int)val, stderr);
#endif

	outPorts[port & 0xff] = val;
}


uns8
in_byte(uns16 port)
{
	int res = 0x1f;
	int highbyte;

	if ((port & 0xff) < 0x10)
		return(0); 
        if ((port & 0x1ff) == HPEN_PORT)
                return(hpen); 

	switch (port & 0xff) {
 	    case KEYBOARD_PORT:  /* keyboard 1 / mouse */
		highbyte = port >> 8;
                if (highbyte ==  0xff)  {
                	res = keyports[8] & 0x1f;
			if (SAMPointerMode && (res == 0x1f)) {
				res = (int)(mouse_in() & 0x1f);
			}
		}

		if (!(highbyte & 0x80)) res &= keyports[7] & 0x1f;
		if (!(highbyte & 0x40)) res &= keyports[6] & 0x1f;
		if (!(highbyte & 0x20)) res &= keyports[5] & 0x1f;
		if (!(highbyte & 0x10)) res &= keyports[4] & 0x1f;
		if (!(highbyte & 0x08)) res &= keyports[3] & 0x1f;
		if (!(highbyte & 0x04)) res &= keyports[2] & 0x1f;
		if (!(highbyte & 0x02)) res &= keyports[1] & 0x1f;
		if (!(highbyte & 0x01)) res &= keyports[0] & 0x1f;
		return (keyboard = (res | (keyboard & 0xe0)));
		break; 

	    case STATUS_PORT:  /* keyboard  2 */
		highbyte = port >> 8;
		res = 0xe0;
		if (!(highbyte & 0x80)) res &= keyports[7] & 0xe0;
		if (!(highbyte & 0x40)) res &= keyports[6] & 0xe0;
		if (!(highbyte & 0x20)) res &= keyports[5] & 0xe0;
		if (!(highbyte & 0x10)) res &= keyports[4] & 0xe0;
		if (!(highbyte & 0x08)) res &= keyports[3] & 0xe0;
		if (!(highbyte & 0x04)) res &= keyports[2] & 0xe0;
		if (!(highbyte & 0x02)) res &= keyports[1] & 0xe0;
		if (!(highbyte & 0x01)) res &= keyports[0] & 0xe0;
		return (((res & 0xe0) | (status_reg & 0x1f)));
		break;

	/* banked memory management */

            case VMPR_PORT:
        	return(vmpr);
        	break;
            case HMPR_PORT:
             	return(hmpr);
             	break;
            case LMPR_PORT:
        	return(lmpr);
		break;
            case CLOCK_PORT:
                return(clock_reg); 
		break;
            case LPEN_PORT:
                return(lpen); 
		break;
            case ATTR_PORT:
                return(attr_reg);
                break;
	/* External memory */
            case HEPR_PORT:
		return(hepr);
		break;
            case LEPR_PORT:
                return(lepr);
                break;
	} /* End switch */
	
        if (((port & 0xff) >= DISK1_PORT &&
             (port & 0xff) < DISK1_PORT + 8) ||
            ((port & 0xff) >=  DISK2_PORT &&
             (port & 0xff) < DISK2_PORT + 8)) 
             	return(fdi_in(port));

        
#ifdef PSEUDO_IO
	if (port == PSEUDO_STDIN)
		return (uns8)getchar();
	else
#endif
			
	return 0xff;
}


/* Any byte in the attribute area with the FLASH bit set has
 * its ink and paper swapped twice/sec.
 * FIXME: this routine is somewhat inefficient at the moment.
 */
void
do_flashing(void)
{
	int i;
	uns8 val;

	if (!GetCfg(flashing))
		return;

	flash_phase = 1 - flash_phase;

        switch (VMPR_SCREEN_MODE) {
	   case MODE_1:
		for (i = 0; i < 768; i++) {
			if ((val = realMem[VMPR_SCREEN_PAGE][6144 + i]) & 0x80)
				attr_write_mode_1(6144 + i, val);
		}
		break;
           case MODE_2:
		for (i = 0; i < 6144; i++) {
			if ((val = realMem[VMPR_SCREEN_PAGE][8192 + i]) & 0x80)
				attr_write_mode_2(8192 + i, val);

          	}
          	break;
           case MODE_3:
           case MODE_4:
                break;
        } /* End switch */

}



/*******************************************
 * Popup dialog handling routines start here
 *
 */

/* Create the popup dialog box */
static void
make_dialog(int *argcnt, char **argvec)
{
	XWMHints xwmh;
	XSizeHints xsh;
	XSetWindowAttributes xswa;
	XTextProperty dlgName;
	char *dlg_name;

	/* first work out the necessary geometries */
	font_ht = theFont->ascent + theFont->descent;
	/* only works with fixed width fonts! */
	char_width = XTextWidth(theFont, " ", 1);

	label_wid = char_width * EF_MAX_VISIBLE;
	label_ht = font_ht;
	entry_wid = (char_width * EF_MAX_VISIBLE) + 4;
	entry_ht = font_ht + 4;
	ok_wid = (char_width * strlen(BUTTON_LABEL)) + 4;
	ok_ht = font_ht + 4;
	dlg_wid = entry_wid + (2 * GUTTER_WID);
	dlg_ht = label_ht + entry_ht + ok_ht + (4 * GUTTER_HT);

	/* window creation */
	dlg_win = XCreateSimpleWindow(dpy, root, dlg_x, dlg_y, dlg_wid, dlg_ht, 
		0, pixels[0], pixels[127]);	/* black & white */
	entry_win = XCreateSimpleWindow(dpy, dlg_win, GUTTER_WID, 
		(2 * GUTTER_HT) + label_ht, entry_wid, entry_ht, 0,
		pixels[0], pixels[127]);
	ok_win = XCreateSimpleWindow(dpy, dlg_win, GUTTER_WID, 
		(3 * GUTTER_HT) + label_ht + entry_ht, ok_wid, ok_ht, 0,
		pixels[0], pixels[127]);


	/* tell the window manager */
	xsh.flags = PSize | PMinSize | PMaxSize | PPosition;
	xsh.min_width  = dlg_wid; xsh.min_height = dlg_ht;
	xsh.max_width  = dlg_wid; xsh.max_height = dlg_ht;

	xwmh.initial_state = NormalState;
	xwmh.input = True;
	xwmh.flags = StateHint | InputHint;

	dlg_name = "xcoupe popup";
	if (XStringListToTextProperty(&dlg_name, 1, &dlgName) == 0) {
		xcoupe_mesg(XCOUPE_ERR, "can't create a TextProperty!");
		return;
	}

	xswa.save_under = True;

	if (GetCfg(priv_cmap))
		XSetWindowColormap(dpy, dlg_win, cmap);
	
	XSelectInput(dpy, dlg_win, ExposureMask | KeyPressMask | EnterWindowMask);
	XSelectInput(dpy, ok_win, ButtonPressMask | ButtonReleaseMask);

	XChangeWindowAttributes(dpy, dlg_win, CWSaveUnder, &xswa);
	XSetTransientForHint(dpy, dlg_win, borderWin);
	XSetWMProperties(dpy, dlg_win, &dlgName, NULL,
		argvec, *argcnt, &xsh, &xwmh, NULL);

	/* subwindows must be mapped & raised */
	XMapRaised(dpy, entry_win);
	XMapRaised(dpy, ok_win);
}


static void
paint_entry_field()
{
	int real_curpos;

	/* draw the string */
	XSetForeground(dpy, theGc, pixels[0]);
	XSetBackground(dpy, theGc, pixels[127]);
	XDrawImageString(dpy, entry_win, theGc, 2, 2 + theFont->ascent, 
		dlg_string + cur_win_pos, strlen(dlg_string + cur_win_pos));

	/* draw the cursor */
	real_curpos = (cursor_pos - cur_win_pos) * char_width + 2;
	XFillRectangle(dpy, entry_win, theGc,
		real_curpos, 2, char_width, entry_ht - 4);

	/* blank the rest of the window */
	XClearArea(dpy, entry_win, real_curpos + char_width, 2,
		(EF_MAX_VISIBLE * char_width) - (real_curpos + char_width + 2),
		font_ht, False);
}


static void
paint_dialog(char *label)
{
	/* draw entry field and button borders */
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, entry_win, theGc, 0, entry_ht - 1, 0, 0);
	XDrawLine(dpy, entry_win, theGc, 0, 0, entry_wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[127]);
	XDrawLine(dpy, entry_win, theGc, entry_wid - 1, 0,
		entry_wid - 1, entry_ht - 1);
	XDrawLine(dpy, entry_win, theGc, entry_wid - 1, entry_ht - 1,
		0, entry_ht - 1);

	XSetForeground(dpy, theGc, pixels[127]);
	XDrawLine(dpy, ok_win, theGc, 0, ok_ht - 1, 0, 0);
	XDrawLine(dpy, ok_win, theGc, 0, 0, ok_wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, ok_win, theGc, ok_wid - 1, 0, ok_wid - 1, ok_ht - 1);
	XDrawLine(dpy, ok_win, theGc, ok_wid - 1, ok_ht - 1, 0, ok_ht - 1);

	/* paint the label & the button label */
	XDrawString(dpy, dlg_win, theGc, GUTTER_WID, GUTTER_HT + theFont->ascent,
		label, strlen(label));
	XDrawString(dpy, ok_win, theGc, 2, 2 + theFont->ascent,
		BUTTON_LABEL, strlen(BUTTON_LABEL));

	paint_entry_field();
}


static void
shadow_in(Window w, int wid, int ht)
{
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, w, theGc, 0, ht - 1, 0, 0);
	XDrawLine(dpy, w, theGc, 0, 0, wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[127]);
	XDrawLine(dpy, w, theGc, wid - 1, 0, wid - 1, ht - 1);
	XDrawLine(dpy, w, theGc, wid - 1, ht - 1, 0, ht - 1);
}

static void
shadow_out(Window w, int wid, int ht)
{
	XSetForeground(dpy, theGc, pixels[127]);
	XDrawLine(dpy, w, theGc, 0, ht - 1, 0, 0);
	XDrawLine(dpy, w, theGc, 0, 0, wid - 1, 0);
	XSetForeground(dpy, theGc, pixels[0]);
	XDrawLine(dpy, w, theGc, wid - 1, 0, wid - 1, ht - 1);
	XDrawLine(dpy, w, theGc, wid - 1, ht - 1, 0, ht - 1);
}


/* Process a keypress in the dialog entry field.  Return a value of 1 if
 * we're finished typing text, 0 otherwise.
 */
static int
process_ef_key(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	switch (ks) {
		case XK_Return:
			return 1;
			break;
		case XK_Escape:
			dlg_string[0] = '\0';
			return 1;
			break;
		case XK_BackSpace: case XK_Delete:
			if (cursor_pos > 0) cursor_pos--;
			dlg_string[cursor_pos] = '\0';
			if (cursor_pos < cur_win_pos) cur_win_pos--;
			break;
		default:
			if (ks < 128 && ks > 31) {
				/* add a character */
				dlg_string[cursor_pos] = (char) ks;
				if (cursor_pos < EF_MAX - 1) cursor_pos++;
				if (cursor_pos >= cur_win_pos + EF_MAX_VISIBLE) cur_win_pos++;
				dlg_string[cursor_pos] = '\0';
			}
			break;
	}

	paint_entry_field();
	XFlush(dpy);

	return 0;
}


/* Pop up the dialog box, and process events until Return is pressed or
 * the OK button is clicked.  Then pop down the dialog box and return the
 * string in the entry field.
 */
char *
dialog_event_loop(char *label)
{
	int finished = 0;
	XEvent xev;
	Window r_ret, w_ret;
	int dx, dy, winx, winy, mask_ret; 

	/* reset everything */
	dlg_string[0] = '\0'; cursor_pos = 0; cur_win_pos = 0;

	/* erase the old label */
	XClearArea(dpy, dlg_win, GUTTER_WID, GUTTER_HT, label_wid, label_ht, False);

	/* move the dialog window to under the pointer */
	if (!XQueryPointer(dpy, root, &r_ret, &w_ret, &dx, &dy,
		&winx, &winy, &mask_ret)) {
		dx = 0; dy = 0;
	} else {
		dx -= dlg_wid / 2; dy -= dlg_ht / 2;
	}
	if (dx < 0) dx = 0;
	if (dy < 0) dy = 0;
	XMoveWindow(dpy, dlg_win, dx, dy);

	XMapRaised(dpy, dlg_win);

	while (!finished) {
		XNextEvent(dpy, &xev);
		switch(xev.type) {
			case Expose:
				if (xev.xexpose.window == mainWin ||
					xev.xexpose.window == borderWin) {
					x_max = HSize - 1; y_max = VSize - 1; x_min = y_min = 0;
					screen_dirty = 1;
					screen_refresh();
				} else {
					paint_dialog(label);
				}
				break;
			case ButtonPress:
				if (xev.xbutton.window == ok_win) {
					shadow_in(xev.xbutton.window, ok_wid, ok_ht);
				}
				break;
			case ButtonRelease:
				if (xev.xbutton.window == ok_win) {
					shadow_out(xev.xbutton.window, ok_wid, ok_ht);
					finished = 1;
				}
				break;
			case KeyPress:
				if (xev.xkey.window == entry_win ||
					xev.xkey.window == dlg_win) {
					finished = process_ef_key((XKeyEvent *)&xev);
				}
				break;
			case EnterNotify:
				if (xev.xcrossing.window == dlg_win)
					XAutoRepeatOn(dpy);
				break;
			default:	/* ignore any other events */
				break;
					
		}
	}

	XUnmapWindow(dpy, dlg_win);
	XFlush(dpy);

	return dlg_string;
}
